#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "haps.h"
#include "md5.h"
#include "list.h"
#include "module.h"

char *
trim (char *string)
{
  int i;

  for (i = strlen (string) - 1; i > 0; i -= 1)
  {
    if ((string[i] == ' ') || (string[i] == '\t') || (string[i] == '\n'))
    {
      string[i] = '\0';
    }
    else
    {
      break;
    }
  }
  return string;
}

char *
extract_word (char *string, int pos, char *delim)
{
  char *temp = (char *) NULL;
  char *left = (char *) NULL;
  char *found = (char *) NULL;
  static char *res = (char *) NULL;
  int current = 1;

  /* first copy original string to keep it ... */
  temp = (char *) malloc (strlen (string) + 1);
  strcpy (temp, string);
  /* now work on the copy */
  left = temp;

  for (found = strstr (left, delim); found; found = strstr (left, delim))
  {
    if (current == pos)
    {
      found[0] = '\0';
      trim (left);
      res = (char *) malloc (strlen (left) + 1);
      strcpy (res, left);
      free (temp);
      return res;
    }
    else
    {
      current += 1;
      left = found + strlen (delim);
    }
  }
  if (current == pos)
  {
    trim (left);
    res = (char *) malloc (strlen (left) + 1);
    strcpy (res, left);
  }
  return res;
}

struct node *
probe (struct node *devices, int params)
{
  char *line;
  char *temp;
  unsigned short proc_id = 0;	/* Processor ID */
  FILE *cpuinfo = NULL;
  size_t len = 0;
  struct node *newdev;
  char buf[0x100];

  devices = nodecondremove (devices, "BUS", "CPU", 1);

  if (!(cpuinfo = fopen (PATH_PROC_CPUINFO, "r")))
  {
    return devices;
  }

  while (getline (&line, &len, cpuinfo) >= 0)
  {
#if defined(__alpha__)
    if (strstr (line, "cpu	"))
    {
#elif defined(__PPC__)
    if (strstr (line, "processor"))
    {
#else
    if (strstr (line, "vendor_id"))
    {
#endif
      proc_id++;

      newdev = nodealloc ();
      nodesetprop (newdev, "BUS", "CPU");
      sprintf (buf, "cpu.%d", proc_id);
      nodesetprop (newdev, "CONN", buf);
      nodesetprop (newdev, "CLASS", "CPU");

      nodesetpropint (newdev, "PROCID", proc_id);
      nodesetprop (newdev, "VENDOR", extract_word (line, 2, ": "));
      nodesetprop (newdev, "MODEL", "");
      nodesetprop (newdev, "CACHE", "");
      nodesetpropint (newdev, "FREQ", 0);
      nodesetprop (newdev, "BUGS", "");
      nodesetpropint (newdev, "BOGOMIPS", 0);
      devices = nodeadd (devices, newdev);
#if defined(__alpha__)
    }
    else if (strstr (line, "cpu model"))
    {
#elif defined(__powerpc__)
    }
    else if (strstr (line, "cpu  "))
    {
#else
    }
    else if (strstr (line, "model name"))
    {
#endif
      nodesetprop (newdev, "MODEL", extract_word (line, 2, ": "));
    }
    else if ((strstr (line, "bogomips")) || (strstr (line, "BogoMIPS")))
    {
      nodesetprop (newdev, "BOGOMIPS", extract_word (line, 2, ": "));
      break;
    }
    else if (strstr (line, "cache size"))
    {
      nodesetprop (newdev, "CACHE", extract_word (line, 2, ":"));
#if defined(__alpha__)
    }
    else if (strstr (line, "cycle frequency [Hz]"))
    {
      nodesetpropint (newdev, "FREQ", atoi (temp = extract_word (line, 2, ":")) / 1000000);
      free (temp);
#elif defined(__powerpc__)
    }
    else if (strstr (line, "clock"))
    {
      nodesetpropint (newdev, "FREQ", atoi (temp = extract_word (line, 2, ":")));
      free (temp);
#else
    }
    else if (strstr (line, "cpu MHz"))
    {
      nodesetpropint (newdev, "FREQ", atoi (temp = extract_word (line, 2, ":")));
      free (temp);
#endif
    }
    else if (strstr (line, "flags"))
    {
      if (strstr (line, "fpu"))
      {
        nodesetprop (newdev, "FPU", "YES");
      }
      if (strstr (line, "mmx"))
      {
        nodesetprop (newdev, "MMX", "YES");
      }
      if (strstr (line, "3dnow"))
      {
        nodesetprop (newdev, "3DNOW", "YES");
      }
    }
    else if (strstr (line, "bug"))
    {
      if (strstr (line, "fdiv") && strstr (line, "yes"))
      {
        nodesetprop (newdev, "BUG_FDIV", "YES");
      }
      else if (strstr (line, "hlt") && strstr (line, "yes"))
      {
        nodesetprop (newdev, "BUG_HLT", "YES");
      }
      else if (strstr (line, "sep") && strstr (line, "yes"))
      {
        nodesetprop (newdev, "BUG_SEP", "YES");
      }
      else if (strstr (line, "f00f") && strstr (line, "yes"))
      {
        nodesetprop (newdev, "BUG_F00F", "YES");
      }
      else if (strstr (line, "coma") && strstr (line, "yes"))
      {
        nodesetprop (newdev, "BUG_COMA", "YES");
      }
    }
  }
  len = 0;
  free (line);
  fclose (cpuinfo);

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
