#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include "md5.h"
#include "list.h"
#include "module.h"

struct node *
probe (struct node *devices, int params)
{
  struct node *newdev;
  unsigned long specid, version;
  char *next, *buf = NULL, *tmp;
  char tmpbuf[4096];
  int fd, bytes = 0;
  int num = 0;
  MD5_CTX context;
  unsigned char digest[16];

  devices = nodecondremove (devices, "BUS", "FIREWIRE", 1);

  fd = open ("/proc/bus/ieee1394/devices", O_RDONLY);
  if (fd >= 0)
  {
    memset (tmpbuf, '\0', 4096);
    while (read (fd, tmpbuf, 4096) > 0)
    {
      buf = realloc (buf, bytes + 4096);
      memcpy (buf + bytes, tmpbuf, 4096);
      bytes += 4096;
      memset (tmpbuf, '\0', 4096);
    }
    close (fd);

    if (buf)
      while (buf && *buf)
      {
	next = strstr (buf + 1, "Node[");
	if (next)
	  *(next - 1) = '\0';
	specid = version = 0;
	if ((tmp = strstr (buf, "Software Specifier ID: ")))
	{
	  specid = strtoul (tmp + 23, NULL, 16);
	  tmp = strstr (buf, "Software Version: ");
	  if (tmp)
	    version = strtoul (tmp + 18, NULL, 16);
	  if (version == 0x010483 && specid == 0x00609e)
	  {
	    newdev = nodealloc ();
	    nodesetprop (newdev, "BUS", "FIREWIRE");
	    sprintf (buf, "fw.%d", num++);
	    nodesetprop (newdev, "CONN", buf);
	    nodesetprop (newdev, "CLASS", "SCSI");
	    nodesetprop (newdev, "DRIVER", "sbp2");

	    tmp = strstr (buf, "Vendor/Model ID:");
	    if (tmp)
	    {
	      char *end;

	      end = strstr (tmp, "\n");
	      if (end)
		*end = '\0';
	      nodesetprop (newdev, "DESC", tmp + 17);
	    }
	    else
	    {
	      nodesetprop (newdev, "DESC",
			   "Generic Firewire Storage Controller");
	    }

	    MD5Init (&context);
	    MD5UpdateS (&context, nodegetprop (newdev, "DESC"));
	    MD5Final (digest, &context);

	    sprintf (buf,
		     "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
		     digest[0], digest[1], digest[2], digest[3], digest[4],
		     digest[5], digest[6], digest[7], digest[8], digest[9],
		     digest[10], digest[11], digest[12], digest[13],
		     digest[14], digest[15]);

	    nodesetprop (newdev, "MD5_ID", buf);

	    devices = nodeadd (devices, newdev);

	  }
	}
	buf = next;
      }
  }

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
