#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "haps.h"
#include "md5.h"
#include "list.h"
#include "module.h"
#include "mem.h"

#define BIOSSTART 0x0F0000
#define BIOSSIZE 0x40000

char *
trim (char *string)
{
  int i;

  for (i = strlen (string) - 1; i > 0; i -= 1)
  {
    if ((string[i] == ' ') || (string[i] == '\t') || (string[i] == '\n'))
    {
      string[i] = '\0';
    }
    else
    {
      break;
    }
  }
  return string;
}

char *
extract_word (char *string, int pos, char *delim)
{
  char *temp = (char *) NULL;
  char *left = (char *) NULL;
  char *found = (char *) NULL;
  static char *res = (char *) NULL;
  int current = 1;

  /* first copy original string to keep it ... */
  temp = (char *) malloc (strlen (string) + 1);
  strcpy (temp, string);
  /* now work on the copy */
  left = temp;

  for (found = strstr (left, delim); found; found = strstr (left, delim))
  {
    if (current == pos)
    {
      found[0] = '\0';
      trim (left);
      res = (char *) malloc (strlen (left) + 1);
      strcpy (res, left);
      free (temp);
      return res;
    }
    else
    {
      current += 1;
      left = found + strlen (delim);
    }
  }
  if (current == pos)
  {
    trim (left);
    res = (char *) malloc (strlen (left) + 1);
    strcpy (res, left);
  }
  return res;
}

extern char *
cut_word (char *string, int start, int len)
{
  char *buffer = NULL;
  int tmp1 = 0;
  int tmp2 = 0;

  buffer = (char *) malloc (len + 1);
  for (tmp1 = start; tmp1 < start + len; tmp1++)
  {
    buffer[tmp2] = string[tmp1];
    tmp2++;
  }
  buffer[tmp2] = '\0';
  return buffer;
}

char *
find_word (char *word, char *file)
{
  FILE *f;
  char *line;
  size_t len = 0;

  if (!(f = fopen (file, "r")))
  {
    return NULL;
  }
  while (getline (&line, &len, f) >= 0)
  {
    if (strstr (line, word))
    {
      if (strstr (word, "\n"))
      {
	fclose (f);
	return cut_word (line, 0, strlen (line) - 1);
      }
      else
      {
	fclose (f);
	return line;
      }
    }
  }
  len = 0;
  free (line);
  fclose (f);
  return NULL;
}

struct node *
probe (struct node *devices, int params)
{
  FILE *f;
  char buf[0x100];
  char ostate[0x100], state[0x100], nstate[0x100], cond[0x100], line[0x100];
  struct node *newdev;
  char *temp, *temp2;

  devices = nodecondremove (devices, "BUS", "MAINBOARD", 1);

  newdev = nodealloc ();
  nodesetprop (newdev, "BUS", "MAINBOARD");
  nodesetprop (newdev, "CONN", "mb.bios");
  nodesetprop (newdev, "CLASS", "BIOS");

  memmd5 (buf, BIOSSTART, BIOSSIZE);

  nodesetprop (newdev, "MD5_ID", buf);
//  nodesetprop(newdev,"MD5_STATE",buf);

  sprintf (state, "UNKNOWN");

  sprintf (buf, "%s/%s", HAPS_PATH, "bioses");
  f = fopen (buf, "rt");
  if (f)
  {
    while (!feof (f))
    {
      fscanf (f, "%[^\n]\n", line);
      if (line[0] == '#')
	continue;

      sscanf (line, "%s %s %s", ostate, cond, nstate);
      if (strncmp (ostate, state, 0x100) == 0)
      {
	if (strncasecmp (cond, "contains|", 9) == 0)
	{
	  if (memcontains (cond + 9, BIOSSTART, BIOSSIZE))
	    sprintf (state, nstate);
	}
      }
    }
  }

  nodesetprop (newdev, "NAME", state);


  nodesetpropint (newdev, "MEM_TOTAL",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("MemTotal",
						 PATH_PROC_MEMINFO), 2,
				      ":")));
  free (temp);
  free (temp2);

  nodesetpropint (newdev, "MEM_FREE",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("MemFree",
						 PATH_PROC_MEMINFO), 2,
				      ":")));
  free (temp);
  free (temp2);

  nodesetpropint (newdev, "MEM_SHARED",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("MemShared",
						 PATH_PROC_MEMINFO), 2,
				      ":")));
  free (temp);
  free (temp2);

  nodesetpropint (newdev, "MEM_BUFFERS",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("Buffers",
						 PATH_PROC_MEMINFO), 2,
				      ":")));
  free (temp);
  free (temp2);

  nodesetpropint (newdev, "MEM_CACHED",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("Cached", PATH_PROC_MEMINFO),
				      2, ":")));
  free (temp);
  free (temp2);

  nodesetpropint (newdev, "MEM_SWAP_TOTAL",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("SwapTotal",
						 PATH_PROC_MEMINFO), 2,
				      ":")));
  free (temp);
  free (temp2);

  nodesetpropint (newdev, "MEM_SWAP_FREE",
		  atoi (temp =
			extract_word (temp2 =
				      find_word ("SwapFree",
						 PATH_PROC_MEMINFO), 2,
				      ":")));
  free (temp);
  free (temp2);

  devices = nodeadd (devices, newdev);

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
