#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <scsi/scsi_ioctl.h>
#include <scsi/scsi.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>

#include "md5.h"
#include "list.h"
#include "module.h"
#include "mem.h"

#define SCSI_ID(s) ((s).dev_id & 0xFF)
#define SCSI_LUN(s) (((s).dev_id >> 8) & 0xFF)
#define SCSI_CHANNEL(s) (((s).dev_id >> 16) & 0xFF)
#define SCSI_HOST(s) (((s).dev_id >> 24) & 0xFF)

#define SCSISCSI_TOP    0
#define SCSISCSI_HOST   1
#define SCSISCSI_VENDOR 2
#define SCSISCSI_TYPE   3

/* Gross */
struct scsi_idlun
{
  unsigned int dev_id;
  unsigned int host_unique_id;
};

void
scsiFindDeviceNames (struct node *device, int forceSeq)
{
  char *devname;
  int letters;
  char devicename[0x100];
  int devcount;
  int host, channel, id, lun, fd;
  struct scsi_idlun scsiId;

  if (strcmp (nodegetprop (device, "CLASS"), "CDROM") == 0)
  {
    devname = "sr";
    letters = 0;
  }
  else if (strcmp (nodegetprop (device, "CLASS"), "TAPE") == 0)
  {
    devname = "st";
    letters = 0;
  }
  else if (strcmp (nodegetprop (device, "CLASS"), "HD") == 0)
  {
    devname = "sd";
    letters = 1;
  }
  else
  {
    devname = "sg";
    letters = 1;
  }

  for (devcount = 0; devcount < 256; devcount++)
  {
    if (!letters)
    {
      sprintf (devicename, "/dev/%s%d", devname, devcount);
    }
    else
    {
      if (devcount < 26)
	sprintf (devicename, "/dev/%s%c", devname, devcount + 'a');
      else
	sprintf (devicename, "/dev/%s%c%c", devname,
		 (devcount / 26 - 1) + 'a', (devcount % 26) + 'a');
    }

    if (access (devicename, O_RDONLY))
    {
      return;
    }

    fd = open (devicename, O_RDONLY | O_NONBLOCK);
    if (fd >= 0)
    {
      ioctl (fd, SCSI_IOCTL_GET_IDLUN, &scsiId);

      /* look for the device we just found */

      nodesetprop (device, "DEVICE", devicename);
      host = atoi (nodegetprop (device, "HOST"));
      channel = atoi (nodegetprop (device, "CHANNEL"));
      id = atoi (nodegetprop (device, "ID"));
      lun = atoi (nodegetprop (device, "LUN"));

      if (host == SCSI_HOST (scsiId) &&
	  channel == SCSI_CHANNEL (scsiId) &&
	  id == SCSI_ID (scsiId) && lun == SCSI_LUN (scsiId))
      {
	nodesetprop (device, "DEVICE", devicename);
	break;
      }

    }
  }
}

static int
loadMissingHosts (int *numMissingHosts)
{
  int num = 0;
  char fn[256];
  int numMissing = 0;
  DIR *dir;
  struct dirent *ent;
  FILE *f;
  char *chptr;
  char *end;

  while (1)
  {
    sprintf (fn, "/proc/scsi/usb-storage-%d", num++);

    if (!(dir = opendir (fn)))
      return numMissing;

    while ((ent = readdir (dir)))
    {
      if (ent->d_name[0] == '.')
	continue;

      strcat (fn, "/");
      strcat (fn, ent->d_name);

      f = fopen (fn, "r");
      if (f)
      {
	while (fgets (fn, sizeof (fn), f))
	{
	  for (chptr = fn; *chptr && isspace (*chptr); chptr++);

	  if (!strncmp (chptr, "Attached:", 9))
	  {
	    chptr = chptr + 9;
	    while (isspace (*chptr))
	      chptr++;

	    /* there is a \n at the end */
	    end = chptr + strlen (chptr) - 2;
	    while (isspace (*end) && end > chptr)
	      *end--;
	    *(end + 1) = '\0';

	    if (*chptr == '0' || !strcasecmp (chptr, "no"))
	    {
	      numMissingHosts[numMissing++] = num - 1;
	    }
	  }
	}

	fclose (f);
      }

    }

    closedir (dir);
  }

  abort ();
}

struct node *
probe (struct node *devices, int params)
{
  int fd;
  char buf[16384];
  char linebuf[80];
  char *scsibuf = NULL;
  char path[64];
  struct stat sb;
  int rc;
  unsigned long bytes = 0;
  int i, state = SCSISCSI_TOP;
  char *start, *chptr, *next, *end;
  int host = -1, channel = -1, id = -1, lun = -1;
  struct node *newdev;
  int missingHosts[256];
  int numMissingHosts = 0;
  MD5_CTX context;
  unsigned char digest[16];

  devices = nodecondremove (devices, "BUS", "SCSI", 1);

  numMissingHosts = loadMissingHosts (missingHosts);

  if (access ("/proc/scsi/scsi", R_OK))
    goto out;

  fd = open ("/proc/scsi/scsi", O_RDONLY);
  if (fd < 0)
    goto out;

  memset (buf, '\0', sizeof (buf));
  while (read (fd, buf, sizeof (buf)) > 0)
  {
    scsibuf = realloc (scsibuf, bytes + sizeof (buf));
    memcpy (scsibuf + bytes, buf, sizeof (buf));
    bytes += sizeof (buf);
    memset (buf, '\0', sizeof (buf));
  }
  close (fd);

  start = scsibuf;
  while (*start)
  {
    chptr = start;
    while (*chptr && *chptr != '\n')
      chptr++;
    *chptr = '\0';
    next = chptr + 1;

    switch (state)
    {
    case SCSISCSI_TOP:
      if (strcmp ("Attached devices: ", start))
      {
	goto out;
      }
      state = SCSISCSI_HOST;
      break;

    case SCSISCSI_HOST:
      if (strncmp ("Host: ", start, 6))
      {
	goto out;
      }

      start = strstr (start, "scsi");
      if (!start)
      {
	goto out;
      }
      start += 4;
      host = strtol (start, NULL, 10);

      start = strstr (start, "Channel: ");
      if (!start)
      {
	goto out;
      }
      start += 9;
      channel = strtol (start, NULL, 10);

      start = strstr (start, "Id: ");
      if (!start)
      {
	goto out;
      }
      start += 4;
      id = strtol (start, NULL, 10);

      start = strstr (start, "Lun: ");
      if (!start)
      {
	goto out;
      }
      start += 5;
      lun = strtol (start, NULL, 10);

      state = SCSISCSI_VENDOR;
      break;

    case SCSISCSI_VENDOR:
      if (strncmp ("  Vendor: ", start, 10))
      {
	goto out;
      }

      start += 10;
      end = chptr = strstr (start, "Model:");
      if (!chptr)
      {
	goto out;
      }

      chptr--;
      while (*chptr == ' ' && *chptr != ':')
	chptr--;
      if (*chptr == ':')
      {
	chptr++;
	*(chptr + 1) = '\0';
	strcpy (linebuf, "Unknown");
      }
      else
      {
	*(chptr + 1) = '\0';
	strcpy (linebuf, start);
      }
      *linebuf = toupper (*linebuf);
      chptr = linebuf + 1;
      while (*chptr)
      {
	*chptr = tolower (*chptr);
	chptr++;
      }

      start = end;		/* beginning of "Model:" */
      start += 7;

      chptr = strstr (start, "Rev:");
      if (!chptr)
      {
	goto out;
      }

      chptr--;
      while (*chptr == ' ')
	chptr--;
      *(chptr + 1) = '\0';

      strcat (linebuf, " ");
      strcat (linebuf, start);

      state = SCSISCSI_TYPE;

      break;

    case SCSISCSI_TYPE:
      if (strncmp ("  Type:", start, 7))
      {
	goto out;
      }

      newdev = nodealloc ();
      nodesetprop (newdev, "BUS", "SCSI");

      if (strstr (start, "Direct-Access"))
      {
	nodesetprop (newdev, "CLASS", "HD");
      }
      else if (strstr (start, "Sequential-Access"))
      {
	nodesetprop (newdev, "CLASS", "TAPE");
      }
      else if (strstr (start, "CD-ROM"))
      {
	nodesetprop (newdev, "CLASS", "CDROM");
      }
      else if (strstr (start, "WORM"))
      {
	nodesetprop (newdev, "CLASS", "CDROM");
      }
      else if (strstr (start, "Scanner"))
      {
	nodesetprop (newdev, "CLASS", "SCANNER");
      }
      else if (strstr (start, "Printer"))
      {				/* WTF? */
	nodesetprop (newdev, "CLASS", "PRINTER");
      }
      else
      {
	nodesetprop (newdev, "CLASS", "OTHER");
      }
      nodesetprop (newdev, "DEVICE", "");
      nodesetpropint (newdev, "HOST", host);
      nodesetpropint (newdev, "CHANNEL", channel);
      nodesetpropint (newdev, "LUN", lun);
      nodesetpropint (newdev, "ID", id);
      sprintf (buf, "scsi.%d.%d.%d.%d", host, channel, lun, id);
      nodesetprop (newdev, "CONN", buf);

      MD5Init (&context);
      MD5UpdateS (&context, nodegetprop (newdev, "DESC"));
      MD5Final (digest, &context);

      sprintf (buf,
	       "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
	       digest[0], digest[1], digest[2], digest[3], digest[4],
	       digest[5], digest[6], digest[7], digest[8], digest[9],
	       digest[10], digest[11], digest[12], digest[13], digest[14],
	       digest[15]);

      nodesetprop (newdev, "MD5_ID", buf);

      nodesetprop (newdev, "DRIVER", "ignore");
      nodesetprop (newdev, "DESC", linebuf);

      i = 0;
      while (1)
      {
	snprintf (path, 63, "/proc/scsi/usb-storage-%d", i);
	if (stat (path, &sb))
	  break;
	snprintf (path, 63, "/proc/scsi/usb-storage-%d/%d", i, host);
	i++;
	fd = open (path, O_RDONLY);
	if (fd < 0)
	  continue;
	memset (buf, '\0', sizeof (buf));
	rc = read (fd, buf, 16384);
	if (rc <= 0)
	  continue;
	buf[rc - 1] = '\0';
	close (fd);
	if (strstr (buf, "Protocol: Uniform Floppy Interface (UFI)"))
	{
	  nodesetprop (newdev, "CLASS", "FLOPPY");
	}
	break;
      }

      for (i = 0; i < numMissingHosts; i++)
      {
	if (missingHosts[i] == host)
	{
	  nodesetprop (newdev, "DETACHED", "1");
	  break;
	}
      }

      scsiFindDeviceNames (newdev, 1);
      devices = nodeadd (devices, newdev);
      state = SCSISCSI_HOST;
    }
    start = next;
  }

out:
  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
