#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>

#include "haps.h"
#include "md5.h"
#include "list.h"
#include "module.h"

struct usbdesc
{
  unsigned int vendorId;
  unsigned int deviceId;
  char *desc;
};

static struct usbdesc *usbDeviceList = NULL;
static int numUsbDevices = 0;

static int devCmp(const void *a, const void *b)
{
	const struct usbdesc *one = a;
	const struct usbdesc *two = b;
	int x, y;

	x = one->vendorId - two->vendorId;
	y = one->deviceId - two->deviceId;
	if (x)
		return x;
	return y;
}

int usbReadDrivers(char *filename)
{
	int fd;
	char *b, *buf, *tmp, *ptr;
	unsigned int vend=0, dev=0;
	struct usbdesc tmpdev;
	char *vendor="";

	if (filename) {
		fd = open(filename, O_RDONLY);
		if (fd < 0)
			return -1;
	} else {
		fd = open(USB_IDS, O_RDONLY);
		if (fd < 0) {
			return -1;
		}
	}
	b = buf = bufFromFd(fd);

	while (*buf) {
		ptr = buf;
		while (*ptr && *ptr != '\n')
			ptr++;
		if (*ptr) {
			*ptr = '\0';
			ptr++;
		}
		if (!strncmp(buf,"# List of known device classes",30))
			break;
		if (*buf == '#') {
			buf = ptr;
			continue;
		}
		if (isalnum(*buf)) {
			tmp = buf;
			while (*tmp && !isspace(*tmp))
				tmp++;
			if (*tmp) {
				*tmp = '\0';
				do
					tmp++;
				while (isspace(*tmp));
			}
			vend = strtol(buf, NULL, 16);
			vendor = tmp;
		}
		if (*buf == '\t') {
			buf++;
			tmp = buf;
			while (*tmp && !isspace(*tmp))
				tmp++;
			if (*tmp) {
				*tmp = '\0';
				do
					tmp++;
				while (isspace(*tmp));
			}
			dev = strtol(buf, NULL, 16);
			if (vend && dev) {
			tmpdev.vendorId = vend;
			tmpdev.deviceId = dev;
			tmpdev.desc =
			    malloc(strlen(tmp) + 2 + strlen(vendor));
			snprintf(tmpdev.desc,
				 strlen(tmp) + 2 + strlen(vendor), "%s %s",
				 vendor, tmp);
			usbDeviceList =
			    realloc(usbDeviceList,
				    (numUsbDevices +
				     1) * sizeof(struct usbdesc));
			usbDeviceList[numUsbDevices] = tmpdev;
			numUsbDevices++;
			}
		}
		buf = ptr;
	}
	free(b);
	qsort(usbDeviceList, numUsbDevices, sizeof(struct usbdesc),
	      devCmp);
	return 0;
}

void usbFreeDrivers()
{
	int x;
	if (usbDeviceList) {
		for (x = 0; x < numUsbDevices; x++) {
			free(usbDeviceList[x].desc);
		}
	}
	free(usbDeviceList);
	usbDeviceList = NULL;
	numUsbDevices = 0;
}

static void usbSearchAndAdd(struct node *usbdev/*, struct device **devlistptr*/)
{
  struct usbdesc *searchDev, key;
	
  key.vendorId = atoi(nodegetprop(usbdev,"VENDORID"));
  key.deviceId = atoi(nodegetprop(usbdev,"DEVICEID"));
  searchDev = bsearch(&key, usbDeviceList, numUsbDevices,
			    sizeof(struct usbdesc), devCmp);

  if (searchDev)
  {
    nodesetprop(usbdev,"DESC",searchDev->desc);
  }

  if (!strcmp(nodegetprop(usbdev,"DESC"),"unknown") && (strcmp(nodegetprop(usbdev,"USBPROD"),"")!=0))
  {
    if (strcmp(nodegetprop(usbdev,"USBMFR"),"")!=0)
    {
      char buf[128];
      snprintf(buf,127,"%s %s",nodegetprop(usbdev,"USBMFR"), nodegetprop(usbdev,"USBPROD"));
      nodesetprop(usbdev,"DESC",buf);
    }
    else
    {
      nodesetprop(usbdev,"DESC",nodegetprop(usbdev,"USBPROD"));
    }
  }
		
}

static void parseDescriptorLine(char *line, struct node *usbdev)
{
  int usbclass,usbsubclass,usbprotocol;

  usbclass=atoi(&line[30]);
  usbsubclass=atoi(&line[44]);
  usbprotocol=atoi(&line[52]);

  nodesetpropint(usbdev,"USBCLASS",usbclass);
  nodesetpropint(usbdev,"USBSUBCLASS",usbsubclass);
  nodesetpropint(usbdev,"USBPROTOCOL",usbprotocol);

  switch (usbclass)
  {
    case 1:
      if (usbsubclass == 02)
      {
        nodesetprop(usbdev,"DRIVER","audio");
        nodesetprop(usbdev,"CLASS","AUDIO");
      }
      else
      {
        nodesetprop(usbdev,"CLASS","OTHER");
      }
      break;
    case 2:
      switch (usbsubclass)
      {
        case 2:
          nodesetprop(usbdev,"CLASS","MODEM");
	  break;
        case 6:
	case 7:
          nodesetprop(usbdev,"CLASS","NETWORK");
	  break;
	default:
          nodesetprop(usbdev,"CLASS","OTHER");
	  break;
      }
      break;
    case 3:
      switch (usbprotocol)
      {
        case 1:
          nodesetprop(usbdev,"DRIVER","keybdev");
          nodesetprop(usbdev,"CLASS","KEYBOARD");
	  break;
        case 2:
          nodesetprop(usbdev,"MOUSETYPE","imps2");
          nodesetprop(usbdev,"XMOUSETYPE","IMPS/2");
          nodesetprop(usbdev,"DRIVER","mousedev");
          nodesetprop(usbdev,"DEVICE","/dev/input/mice");
          nodesetprop(usbdev,"CLASS","MOUSE");
	  break;
        default:
          nodesetprop(usbdev,"CLASS","OTHER");
	  break;
      }
      break;
    case 7:
      nodesetprop(usbdev,"CLASS","PRINTER");
      break;
    case 8:
      switch (usbsubclass)
      {
        case 2:
          nodesetprop(usbdev,"DRIVER","usb-storage");
          nodesetprop(usbdev,"CLASS","CDROM");
	  break;
        case 3:
          nodesetprop(usbdev,"CLASS","TAPE");
	  break;
        case 4:
          nodesetprop(usbdev,"DRIVER","usb-storage");
          nodesetprop(usbdev,"CLASS","FLOPPY");
	  break;
        case 5:
	  /* Could be a M/O drive. Could be a floppy drive. */
          nodesetprop(usbdev,"CLASS","OTHER");
	  break;
        case 6:
          nodesetprop(usbdev,"DRIVER","usb-storage");
          nodesetprop(usbdev,"CLASS","HD");
	  break;
      }
    default:
      nodesetprop(usbdev,"CLASS","OTHER");
      break;
  }
}

static void parseStringDescriptorLine(char *line, struct node *usbdev)
{
  int x;
  char *tmp;
  char buf[0x100];

  if ((tmp = strcasestr(line, "product")) != NULL)
  {
    strcpy(buf,tmp+8);

    for (x=0;buf[x];x++)
      if (buf[x]=='\n')
        buf[x]='\0';

    for (x-=2; x>=0; x--)
    {
      if (isspace(buf[x]))
        buf[x] = '\0';
      else
	break;
    }
    nodesetprop(usbdev,"USBPROD",buf);
  }

  if ((tmp = strcasestr(line, "manufacturer")) != NULL)
  {
    strcpy(buf,tmp+13);

    for (x=0;buf[x];x++)
      if (buf[x]=='\n')
        buf[x]='\0';

    for (x-=2; x>=0; x--)
    {
      if (isspace(buf[x]))
        buf[x] = '\0';
      else
	break;
    }
    nodesetprop(usbdev,"USBMFR",buf);
  }
}


void setusbmd5(struct node *device)
{
  MD5_CTX context;
  unsigned char digest[16];
  char buf[0x100];

  MD5Init (&context);
  MD5UpdateS(&context,nodegetprop(device,"USBPROD"));
  MD5UpdateS(&context,nodegetprop(device,"USBMFR"));
  MD5UpdateS(&context,nodegetprop(device,"VENDORID"));
  MD5UpdateS(&context,nodegetprop(device,"DEVICEID"));
  MD5Final (digest, &context);
  
  sprintf(buf,"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
    digest[0],digest[1],digest[2],digest[3],
    digest[4],digest[5],digest[6],digest[7],
    digest[8],digest[9],digest[10],digest[11],
    digest[12],digest[13],digest[14],digest[15]);
  nodesetprop(device,"MD5_ID",buf);
//  nodesetprop(device,"MD5_STATE",buf);
}

struct node *probe(struct node *devices,int params)
{
  FILE *usbdevicelist;
  char line[0x100],buf[0x100];
  struct node *usbdev=NULL,*tmpdev=NULL;

  devices=nodecondremove(devices,"BUS","USB",1);

  if (!usbDeviceList)
  {
    usbReadDrivers(NULL);
  }

  usbdevicelist = fopen("/proc/bus/usb/devices", "r");
  if (usbdevicelist != NULL)
  {
    while (fgets(line, 255, usbdevicelist))
    {
      switch (line[0])
      {
	case 'T':
          if (usbdev != NULL)
	  {
	    usbSearchAndAdd(usbdev);

            devices=nodeadd(devices,usbdev);
	  }
          usbdev=nodealloc();
          nodesetprop(usbdev,"BUS","USB");
          nodesetprop(usbdev,"DESC","UNKNOWN");
          nodesetprop(usbdev,"DRIVER","UNKNOWN");
          nodesetpropint(usbdev,"USBBUS",atoi(&line[8]));
          nodesetpropint(usbdev,"USBLEVEL",atoi(&line[15]));
          nodesetpropint(usbdev,"USBPORT",atoi(&line[31]));
	  sprintf(buf,"usb.%d.%d.%d",atoi(&line[8]),atoi(&line[15]),atoi(&line[31]));
          nodesetprop(usbdev,"CONN",buf);

	  break;
	case 'I':
	  if (atoi(line+8) > 0)
	  {
	    if (usbdev != NULL)
	    {
              tmpdev=nodeclone(usbdev);
	      usbSearchAndAdd(usbdev);
              devices=nodeadd(devices,usbdev);
	      usbdev=tmpdev;
	    }
	  }
	  parseDescriptorLine(line,usbdev);
	  break;
	case 'P':
          nodesetpropint(usbdev,"VENDORID",strtol(&line[11], NULL, 16));
          nodesetpropint(usbdev,"DEVICEID",strtol(&line[23], NULL, 16));
	case 'S':
	  parseStringDescriptorLine(line,usbdev);
	  break;
	default:
	  break;
      }
    }
    if (usbdev != NULL)
    {
      usbSearchAndAdd(usbdev);
      devices=nodeadd(devices,usbdev);
    }
    fclose(usbdevicelist);
  }

  usbFreeDrivers();

  return devices;
}


#ifdef HAPS_MODULES
struct modulehandler *_init_module(void)
{
  struct modulehandler *mh;
  mh=malloc(sizeof(*mh));
  mh->probe=probe;
  return mh;
}
#endif
