#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
		     
#include "haps.h"
#include "md5.h"
#include "list.h"
#include "module.h"

int
exists (char *filename)
{
  struct stat s;
  return !stat (filename, &s);
}

struct node *
probe (struct node *devices, int params)
{
  struct node *curdev;
  char *xfree4 = "XFree86", *xvesa4 = "vesa";
  const char *xpath = XPATH;
  char *p;
  char xmodule[32];
  char xserver[32];
  struct utsname utsbuf;
  char buf[0x100];
  int fd;
  
  int rescanned = 0;
  const char *driver[] = { "3DLabs", "Mach64", "Mach32", "Mach8", "AGX",
    "P9000", "S3 ViRGE", "S3V", "S3", "W32",
    "8514", "I128", "SVGA", xfree4, NULL
  };
  const char *server[] =
    { driver[0], driver[1], driver[2], driver[3], driver[4],
    driver[5], "S3V", driver[7], driver[8], driver[9],
    driver[10], driver[11], driver[12], xfree4, NULL
  };

  for (curdev = devices; curdev != NULL; curdev = curdev->next)
  {
    if (strcmp (nodegetprop (curdev, "CLASS"), "VIDEO") == 0)
    {

      if (nodegetprop (curdev, "DRIVER"))
      {
	if (!strncasecmp (nodegetprop (curdev, "DRIVER"), "Card:", 5))	/* RedHat Cards-DB */
	{
	  /* Kudzu "Cards" format */
	  FILE *cardsdb;
	  char xfree3server[128];
	  memset (xfree3server, 0, sizeof (xfree3server));
	  if ((cardsdb = fopen (CARDSDB, "r")) != NULL)
	  {
	    /* Try to find Server and Module in /usr/share/kudzu/Cards */
	    char buffer[1024];
	    char searchfor[128];
	    int found = 0;
	    memset (searchfor, 0, sizeof (searchfor));
	    p=nodegetprop (curdev, "DRIVER");
	    sscanf (&p[5], "%127[^\r\n]", searchfor);
	    while (!found && !feof (cardsdb) && fgets (buffer, 1024, cardsdb))
	    {
	      char sfound[128];
	      memset (sfound, 0, sizeof (sfound));
	      if (strncasecmp (buffer, "NAME ", 5) ||
		  (sscanf (&buffer[5], "%127[^\r\n]", sfound) != 1) ||
		  strcasecmp (sfound, searchfor))
		continue;
	      while (!feof (cardsdb) && fgets (buffer, 1024, cardsdb))
	      {
		if (buffer[0] < 32)
		  break;	/* End-of-line */
		if (!strncasecmp (buffer, "SERVER ", 7))
		{
		  char x[20] = "";
		  if (sscanf (&buffer[7], "%19s", x) == 1)
		  {
		    char fullpath[128];
		    char *xf[2] = { "", "XF86_" };
		    int i;
		    for (i = 0; i < 2; i++)
		    {
		      sprintf (xserver, "%s%.24s", xf[i], x);
		      sprintf (fullpath, "%.90s%.32s", xpath, xserver);
		      if (exists (fullpath))
		      {
			strncpy (xfree3server, xserver,
				 sizeof (xfree3server));
			break;	/* for */
		      }
		    }
		  }
		}
		else if (!strncasecmp (buffer, "DRIVER ", 7))
		{
		  char fullpath[128];
		  sscanf (&buffer[7], "%31s", xmodule);
		  sprintf (fullpath, XMODPATH "%.31s_drv.o", xmodule);
		  if (exists (fullpath))
		  {
		    nodesetprop (curdev, "XMODULE", xmodule);
		  }
		}
		else if (!strncasecmp (buffer, "SEE ", 4) && rescanned < 10)
		{		/* rescan Cards-DB for other server */
		  fseek (cardsdb, 0L, SEEK_SET);
		  ++rescanned;
		  memset (searchfor, 0, sizeof (searchfor));
		  sscanf (&buffer[4], "%127[^\r\n]", searchfor);
		  break;	/* Continue with outer while() */
		}
	      }
	    }
	    fclose (cardsdb);
	  }
	  if (nodegetprop (curdev, "XMODULE")!=NULL || nodegetprop (curdev, "XSERVER")!=NULL || *xfree3server)	/* (Partial) Success */
	  {
	    if (nodegetprop (curdev, "XSERVER")==NULL)
	    {
	      if (*xfree3server && nodegetprop (curdev, "XMODULE")==NULL)
	        nodesetprop (curdev, "XSERVER",xfree3server);
	      else
	        nodesetprop (curdev, "XSERVER",xfree4);
	    }
	    if (nodegetprop (curdev, "XMODULE")==NULL)
	      nodesetprop (curdev, "XMODULE",xvesa4);
	    continue;
	  }
	}
	/* Card not found in Cards database -> Try to guess from description */
	{
	  int i;
	  for (i = 0; driver[i] != NULL; i++)
	  {
	    if (strstr (nodegetprop (curdev, "DRIVER"), driver[i]))
	    {
	      char *xpos;
	      if ((xpos = strstr (nodegetprop (curdev, "DRIVER"), xfree4)) != NULL)	/* Check for XFree 4 */
	      {
		char xm[32] = "";
	        nodesetprop (curdev, "XSERVER",xfree4);
		if (sscanf (xpos, "XFree86(%30[^)])", xm) == 1)
	          nodesetprop (curdev, "XMODULE",xm);
		else
	          nodesetprop (curdev, "XMODULE",xvesa4);
	      }
	      else
	      {
		char xserver[32];
		char fullpath[128];
		char *xf[2] = { "", "XF86_" };
		int j;
		for (j = 0; j < 2; j++)
		{
		  sprintf (xserver, "%s%.24s", xf[j], server[i]);
		  sprintf (fullpath, "%.90s%.32s", xpath, xserver);
		  if (exists (fullpath))
		  {
	            nodesetprop (curdev, "XSERVER",xserver);
		    break;	/* for */
		  }
		}
	      }
	    }
	  }
	}
      }
      /* Special options required? */
      if (nodegetprop (curdev, "DESC"))
      {
	/* Handle special cards that require special options */
	if (strstr (nodegetprop (curdev, "DESC"), "Trident")
	    || strstr (nodegetprop (curdev, "DESC"), "TGUI")
	    || strstr (nodegetprop (curdev, "DESC"), "Cirrus")
	    || strstr (nodegetprop (curdev, "DESC"), "clgd"))
	{
	  if (!strcmp (nodegetprop (curdev, "XSERVER"), xfree4))
            nodesetprop (curdev, "XOPTS","-depth 16");
	  else
            nodesetprop (curdev, "XOPTS","-bpp 16");
	}
	else if (strstr (nodegetprop (curdev, "DESC"), "Savage 4"))	/* S3 Inc.|Savage 4 */
	{
	  if (!strcmp (nodegetprop (curdev, "XSERVER"), xfree4))
            nodesetprop (curdev, "XOPTS","-depth 32");
	  else
            nodesetprop (curdev, "XOPTS","-bpp 32");
	}
      }
      /* Fallback values */
      if (nodegetprop (curdev, "XSERVER")==NULL)
        nodesetprop (curdev, "XSERVER",xfree4);
      if (nodegetprop (curdev, "XMODULE")==NULL)
        nodesetprop (curdev, "XMODULE",xvesa4);
	
      if (strstr (nodegetprop (curdev, "XMODULE"), "nv"))
      {
        uname (&utsbuf);
        snprintf (buf, 255, "/lib/modules/%s/misc/nvidia.o", utsbuf.release);
        if ((fd=open (buf, O_RDONLY)) >= 0)
        {
          nodesetprop (curdev, "XMODULE","nvidia");
	  close(fd);
        } else
	{
          snprintf (buf, 255, "/lib/modules/%s/misc/nvidia.o.gz", utsbuf.release);
          if ((fd=open (buf, O_RDONLY)) >= 0)
          {
            nodesetprop (curdev, "XMODULE","nvidia");
	    close(fd);
          }
	}
      }
    }
  }

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
