#include "list.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

struct node *nodealloc()
{
  struct node *n=(struct node*)malloc(sizeof(struct node));

  n->prev=NULL;
  n->next=NULL;
  n->properties=NULL;
  
  return n;
}


struct node *nodeadd(struct node *set,struct node *n)
{
  struct node *cur,*prev;
  if (set==NULL)return n;

  for (cur=set;cur!=NULL;cur=cur->next)
  {
    prev=cur;
  };
  prev->next=n;
  n->prev=prev;
  
  return set;
}

struct node *nodedel(struct node *n)
{
  struct property *p;
  struct node *ret=NULL;
  
  if (n==NULL)return NULL;
  
  if (n->prev!=NULL)n->prev->next=n->next;
  if (n->next!=NULL)n->next->prev=n->prev;

  if (n->prev!=NULL)ret=n->prev;
  if (n->next!=NULL)ret=n->next;
  
  p=n->properties;
  while ((p=propertydel(p)));
  free(n);
  
  return ret;
}

struct node *nodefree(struct node *n)
{
  while ((n=nodedel(n)));
  return NULL;
}

void nodesetprop(struct node *n,char *name,char *value)
{
  struct property *nprop;

  nprop=propertyalloc();

  if (propertyset(nprop,name,value)==-1)
  {
    free(nprop);
    return;
  };
  n->properties=propertyadd(n->properties,nprop);
}

void nodesetpropint(struct node *n,char *name,int value)
{
  char buf[0x10];
  sprintf(buf,"%d",value);
  nodesetprop(n,name,buf);
}

void nodesetprophint(struct node *n,char *name,unsigned int value)
{
  char buf[0x10];
  sprintf(buf,"%x",value);
  nodesetprop(n,name,buf);
}

char *nodegetprop(struct node *n,char *name)
{
  struct property *prop;

  for (prop=n->properties;prop!=NULL;prop=prop->next)
  {
    if (strcmp(prop->name,name)==0)break;
  }
  if (prop==NULL)return NULL; else return prop->value;
}



int shenv(char *cmd,char **env)
{
  char *argv[3];
  int ppid;
  
  
  ppid=fork();
  if (ppid==-1)return -1;
  if (ppid==0)
  {
    argv[0]="/bin/sh";
    argv[1]="-c";
    argv[2]=cmd;
    argv[3]=NULL;
    execve("/bin/sh",argv,env);
    exit(0);
  } else
  wait(NULL);

  return 0;
}

int nodesh(char *cmd,struct node *devices,int loopmode)
{
  char **env;
  struct node *curdev,*newdev,*retdevs=NULL;
  struct property *prop;
  int size,i;
  
  for (curdev=devices;curdev!=NULL;curdev=curdev->next)
  {
    newdev=nodealloc();
    retdevs=nodeadd(retdevs,newdev);
    size=0;
    for (prop=curdev->properties;prop!=NULL;prop=prop->next) size++;
    env=(char**)malloc((size+1)*sizeof(char*));
    env[size]=NULL;
    i=0;
    for (prop=curdev->properties;prop!=NULL;prop=prop->next)
    {
      asprintf(&env[i++],"%s=%s",prop->name,prop->value);
    }
    shenv(cmd,env);
    for (i=0;env[i]!=NULL;i++) free(env[i]);
    free(env);
    if (loopmode!=1)break; // FORFIRST
  }
  return 0;
}

struct node *nodecondremove(struct node *devices,char *name,char *value,int cond)
{
  struct node *curdev;
  char *val;
  int del;

  curdev=devices;
  while (curdev!=NULL)
  {
    val=nodegetprop(curdev,name);
    if (val==NULL)val="";
    del=0;
    if (cond==0) // ALL
    {
      del=1;
    } else
    if (cond==1) // EQ
    {
      if (strcmp(val,value)==0)del=1;
    } else
    if (cond==2) // NEQ
    {
      if (strcmp(val,value)!=0)del=1;
    }
    
    if (del==1)
    {
      if (devices==curdev)
        curdev=devices=nodedel(curdev);
      else
        curdev=nodedel(curdev);
    } else
    {
      curdev=curdev->next;
    }
  }
  return devices;
}

int nodesave(char *filename,struct node *devices)
{
  struct node *curdev;
  struct property *prop;
  FILE *f;

  if (!(f=fopen(filename,"wt")))return -1;
  
  for (curdev=devices;curdev!=NULL;curdev=curdev->next)
  {
    fprintf(f,"NODEBEGIN\n");
    for (prop=curdev->properties;prop!=NULL;prop=prop->next)
    {
      if (prop->name==NULL)continue;
      if (prop->value==NULL)prop->value="";
      fprintf(f,"%s=\"%s\"\n",prop->name,prop->value);
    }
    fprintf(f,"NODEEND\n\n");
  }
  
  fclose(f);
  return 0;
}

struct node *nodeload(char *filename)
{
  struct node *newdev=NULL,*devices=NULL;
  FILE *f;
  char buf[0x100],name[0x100],value[0x100];
  int indev=0;

  if (!(f=fopen(filename,"rt")))return NULL;
  while (!feof(f))
  {
    fscanf(f,"%[^\n]\n",buf);
    if (strncmp(buf,"NODEBEGIN",12)==0)
    {
      newdev=nodealloc();
      devices=nodeadd(devices,newdev);
      indev=1;
    } else
    if (strncmp(buf,"NODEEND",10)==0) indev=0; else
    if (indev)
    {
      sscanf(buf,"%[^=]=\"%[^\"]\"",name,value);
      nodesetprop(newdev,name,value);
    }
  }
  fclose(f);
  return devices;
}

struct node *nodemerge(struct node *devices1,struct node *devices2,char *uniq)
// dodaje do 1 urzdzenia z 2 pod warunkiem, e jeszcze nie wystpuj
{
  struct node *curdev1,*curdev2,*newdev;
  struct property *prop;
  char *val1,*val2;
  
  for (curdev1=devices2;curdev1!=NULL;curdev1=curdev1->next)
  {
    val1=nodegetprop(curdev1,uniq);
    if (val1==NULL)continue; // pomijamy urzdzenia bez uniq

    for (curdev2=devices1;curdev2!=NULL;curdev2=curdev2->next)
    {
      val2=nodegetprop(curdev2,uniq);
      if ((val2!=NULL)&&(strcmp(val1,val2)==0))break;
    }
    if (curdev2==NULL) // nie znaleziono
    {
      newdev=nodealloc();
      devices1=nodeadd(devices1,newdev);
      for (prop=curdev1->properties;prop!=NULL;prop=prop->next)
      {
        nodesetprop(newdev,prop->name,prop->value);
      }
    }
  }

  return devices1;
}

struct node *nodecopy(struct node *devices)
{
  struct node *curdev,*newdev,*retdevs=NULL;
  struct property *prop;

  for (curdev=devices;curdev!=NULL;curdev=curdev->next)
  {
    newdev=nodealloc();
    retdevs=nodeadd(retdevs,newdev);

    for (prop=curdev->properties;prop!=NULL;prop=prop->next)
    {
      nodesetprop(newdev,prop->name,prop->value);
    }
  }
  return retdevs;
}

struct node *nodeclone(struct node *curdev)
{
  struct node *newdev;
  struct property *prop;

  if (curdev==NULL)return NULL;

  newdev=nodealloc();
  for (prop=curdev->properties;prop!=NULL;prop=prop->next)
  {
    nodesetprop(newdev,prop->name,prop->value);
  }
  return newdev;
}

struct node *nodediff(struct node *devices1,struct node *devices2,char *uniq,char *field)
// zwraca urzdzenia wystpujce w 1, a nie wystpujce w 2
// zwraca urzdzenia wystpujce w 1, i wystpujce w 2, ale posiadajce rne field
{
  struct node *curdev1,*curdev2,*newdev,*retdevs=NULL;
  char *val1,*val2;
  char *val3,*val4;

  for (curdev1=devices1;curdev1!=NULL;curdev1=curdev1->next)
  {
    val1=nodegetprop(curdev1,uniq);
    if (val1==NULL)continue;
    val3=nodegetprop(curdev1,field);
    if (val3==NULL)val3="";
    val4=NULL;

    for (curdev2=devices2;curdev2!=NULL;curdev2=curdev2->next)
    {
      val2=nodegetprop(curdev2,uniq);
      if (val2==NULL)continue;

      if ((strcmp(val1,val2)==0))
      {
        val4=nodegetprop(curdev1,field);
        if (val4==NULL)val4="";
	break;
      }
    }

    if (val4!=NULL) // isnieja oba
    {
      if (strcmp(val3,val4)!=0)
      {
        newdev=nodeclone(curdev1);
        retdevs=nodeadd(retdevs,newdev);
      }
    }
    
    if (curdev2==NULL) // nie znaleziono na 2 licie
    {
      newdev=nodeclone(curdev1);
      retdevs=nodeadd(retdevs,newdev);
    }
  }

  return retdevs;
}

///////////////////////////////////////////////////////////////////////////
struct property *propertyalloc()
{
  struct property *n=(struct property*)malloc(sizeof(struct property));

  n->prev=NULL;
  n->next=NULL;

  n->name=NULL;
  n->value=NULL;
  
  return n;
}

int propertyset(struct property *prop,char *name,char *value)
{
  if (name==NULL)return -1;
  if (value==NULL)return -1;
  
  if (prop->name!=NULL) free(prop->name);
  if (prop->value!=NULL) free(prop->value);

  prop->name=(char *)malloc(strlen(name)+1);
  prop->value=(char *)malloc(strlen(value)+1);
  
  strcpy(prop->name,name);
  strcpy(prop->value,value);
  return 0;
}

struct property *propertyadd(struct property *set,struct property *nprop)
{
  struct property *cur,*prev=NULL;

  if (set==NULL)return nprop;

  for (cur=set;cur!=NULL;cur=cur->next)
  {
    if ((cur->name!=NULL) && (strcmp(cur->name,nprop->name)==0)) break;
    prev=cur;
  }
  if (cur==NULL)
  {
    if (prev!=NULL)prev->next=nprop;
    nprop->prev=prev;
  } else
  {
    propertyset(cur,nprop->name,nprop->value);
    free(nprop);
  }
  
  return set;
}

struct property *propertysearch(char *name,struct property *set)
{
  struct property *cur;

  for (cur=set;cur!=NULL;cur=cur->next)
  {
    if ((cur->name!=NULL) && (strcmp(cur->name,name)==0)) break;
  }
  return cur;
}

struct property *propertydel(struct property *cur)
{
  struct property *ret=NULL;

  if (cur!=NULL)
  {
    if ((cur->prev!=NULL)&&(cur->prev->next!=NULL))cur->prev->next=cur->next;
    if ((cur->next!=NULL)&&(cur->next->prev!=NULL))cur->next->prev=cur->prev;

    if (cur->prev!=NULL) ret=cur->prev;
    if (cur->next!=NULL) ret=cur->next;

    if (cur->name!=NULL) free(cur->name);
    if (cur->value!=NULL) free(cur->value);
  }

  free(cur);
  
  return ret;
}
